from twisted.internet.protocol import DatagramProtocol
from twisted.internet.task import LoopingCall
from twisted.internet import reactor

from collections import Counter, defaultdict, deque
import heapq
from time import time
import argparse
import ConfigParser

class StatsdServer(DatagramProtocol):

    def __init__(self, config):
        self.config = config
        self.backends = []
        self.type_handlers = {
            'ms': self.handleTimer,
            'c': self.handleCounter,
            'g': self.handleGauge,
        }
        self.timers = defaultdict(list)
        self.timers_sum = Counter()
        self.counters = Counter()
        self.gauges_sum = Counter()
        self.gauges_count = Counter()
        self.parseRepeaters()

        # Setup flush routine
        self.flush_call = LoopingCall(self.handleFlush)
        self.flush_call.start(float(config.get("YAStatsd", "FlushInterval")))

    def addBackend(self, backend):
        self.backends.append(backend)

    def datagramReceived(self, data, (host, port)):
        self.repeat(data)
        metrics = data.split('\n')
        for metric in metrics:
            if metric != '':
                self.metricReceived(metric, (host, port))

    def metricReceived(self, metric, (host, port)):
        met_split = metric.split('|')
        if len(met_split) == 2:
            event, event_type = met_split
            sampling = '@1'
        elif len(met_split) == 3:
            event, event_type, sampling = met_split
        else:
            raise ValueError('Invalid number of "|"s, found %d'
                % (len(met_split)+1))
        sampling = float(sampling[1:])
        event_name, event_str_val = event.split(':')
        event_val = float(event_str_val)
        self.type_handlers[event_type](event_name, event_type,
            event_val, sampling)

    def handleTimer(self, event_name, event_type, event_val, sampling):
        # Sampling is meaningless for a timer
        if sampling != 1:
            raise ValueError('Sampling specified for a timer type')
        
        heapq.heappush(self.timers[event_name], event_val)
        self.timers_sum[event_name] += event_val

    def handleCounter(self, event_name, event_type, event_val, sampling):
        if sampling != 1:
            self.counters[event_name] += event_val * (1 / sampling)
        else:
            self.counters[event_name] += event_val

    def handleGauge(self, event_name, event_type, event_val, sampling):
        self.gauges_sum[event_name] += event_val
        self.gauges_count[event_name] += 1

    def handleFlush(self):
        timeval = int(time())

        for backend in self.backends:
            backend.handleFlush(self, timeval)

        self.timers = defaultdict(list)
        self.timers_sum = Counter()
        self.gauges_sum = Counter()
        self.gauges_count = Counter()
        self.counters = Counter()

    def parseRepeaters(self):
        self.repeaters = []
        if (self.config.has_option("YAStatsd", "repeater") and
                len(self.config.get("YAStatsd", "repeater")) > 0):
            for rep in self.config.get("YAStatsd", "repeater").split(','):
                if ":" in rep:
                    addr = rep.split(":")
                    self.repeaters.append(
                        (addr[0].strip(), int(addr[1].strip())))
                else:
                    self.repeaters.append(
                        (rep.strip(),
                         int(self.config.get("YAStatsd", "Port"))))

    def repeat(self, data):
        for rep in self.repeaters:
            self.transport.write(data, rep)


def run_server(config):
    ss = StatsdServer(config)

    from backends.graphite import GraphiteBackend
    gb = GraphiteBackend(config)
    ss.addBackend(gb)
    reactor.connectTCP(
        config.get("Graphite", "Host"),
        int(config.get("Graphite", "Port")),
        gb)

    reactor.listenUDP(int(config.get("YAStatsd", "Port")), ss)
    reactor.run()


def main():
    argparser = argparse.ArgumentParser()
    argparser.add_argument("config", type=str,
        help="Path to configuration file")
    args = argparser.parse_args()

    config = ConfigParser.SafeConfigParser()
    config.read(args.config)
    run_server(config)

if __name__ == '__main__':
    main()



from twisted.internet.protocol import ClientFactory, Protocol
from twisted.internet.task import LoopingCall
from twisted.internet import reactor

from collections import deque
import heapq
import math


class GraphiteConnection(Protocol):

    def __init__(self, factory):
        self.factory = factory

    def connectionMade(self):
        self.factory.connections.append(self)

    def connectionLost(self, reason):
        self.factory.connections.remove(self)


class GraphiteBackend(ClientFactory):

    def __init__(self, config):
        self.config = config
        self.connections = []

    def buildProtocol(self, addr):
        return GraphiteConnection(self)

    def clientConnectionLost(self, connector, reason):
        print 'Connection to graphite lost, Reason:', reason
        conn_retry = int(self.config.get("Graphite", "ConnRetryInterval"))
        print 'Retrying in %d seconds' % conn_retry
        reactor.callLater(conn_retry, connector.connect)

    def clientConnectionFailed(self, connector, reason):
        print 'Connection to graphite failed, Reason:', reason
        conn_retry = int(self.config.get("Graphite", "ConnRetryInterval"))
        print 'Retrying in %d seconds' % conn_retry
        reactor.callLater(conn_retry, connector.connect)

    def handleFlush(self, stats, time):
        self.flushTimers(stats, time)
        self.flushCounters(stats, time)
        self.flushGauges(stats, time)

    def flushTimers(self, stats, time):
        msgs = deque()
        timer_prefix = self.config.get("Graphite", "TimerPrefix")
        percent_thresholds = [int(x) for x in
            self.config.get("YAStatsd", "PercentThresholds").split(',')]
        template_args = {
            'prefix': timer_prefix,
            'time': time
        }
        pct_template = '%(prefix)s.%(timer_name)s.%(stat_type)s_%(pct)d %(val)f\n'
        val_template = '%(prefix)s.%(timer_name)s.%(stat_type)s %(val)f %(time)d\n'
        stat_types = ('sum', 'mean', 'upper', 'count')
        for timer_name, timer_vals in stats.timers.items():
            template_args['timer_name'] = timer_name

            for pct_threshold in percent_thresholds:
                template_args['pct'] = pct_threshold
                pct_elem_cnt = int(math.ceil(len(timer_vals) * pct_threshold * .01))
                pct_vals = heapq.nsmallest(pct_elem_cnt, timer_vals)
                pct_sum = sum(pct_vals)
                template_args['sum'] = pct_sum
                template_args['mean'] = pct_sum / float(len(pct_vals))
                template_args['upper'] = pct_vals[-1]
                template_args['count'] = len(pct_vals)

                for stat_type in stat_types:
                    msgs.append(pct_template % dict(
                            template_args,
                            stat_type=stat_type,
                            val=template_args[stat_type]))

            timer_sum = stats.timers_sum[timer_name]
            template_args['sum'] = timer_sum
            template_args['mean'] = timer_sum / float(len(timer_vals))
            template_args['upper'] = heapq.nlargest(1, timer_vals)[0]
            template_args['count'] = len(timer_vals)

            for stat_type in stat_types:
                msgs.append(val_template % dict(
                    template_args,
                    stat_type=stat_type,
                    val=template_args[stat_type]))

        for conn in self.connections:
            for msg in msgs:
                conn.transport.write(msg)
            conn.transport.write('%(prefix)s.%(name)s %(val)d %(time)d\n' %
                {
                    'prefix': self.config.get("Graphite", "YAStatsdPrefix"),
                    'name': 'timers.count',
                    'val': len(stats.timers),
                    'time': time
                })

    def flushCounters(self, stats, time):
        msgs = deque()
        for name, value in stats.counters.items():
            msgs.append('%(prefix)s.%(name)s %(value)f %(time)d\n' %
                { 
                    'prefix': self.config.get("Graphite", "CounterPrefix"),
                    'name': name,
                    'value': value / float(self.config.get("YAStatsd", "FlushInterval")),
                    'time': time
                })
        for conn in self.connections:
            for msg in msgs:
                conn.transport.write(msg)
            conn.transport.write('%(prefix)s.%(name)s %(val)d %(time)d\n' %
                {
                    'prefix': self.config.get("Graphite", "YAStatsdPrefix"),
                    'name': 'counters.count',
                    'val': len(stats.counters),
                    'time': time
                })

    def flushGauges(self, stats, time):
        msgs = deque()
        for name, value in stats.gauges_sum.items():
            avg = value / float(stats.gauges_count[name])
            msgs.append('%(prefix)s.%(name)s %(avg)f %(time)d\n' %
                {
                    'prefix': self.config.get("Graphite", "GaugePrefix"),
                    'name': name,
                    'avg': avg,
                    'time': time
                })
        for conn in self.connections:
            for msg in msgs:
                conn.transport.write(msg)
            conn.transport.write('%(prefix)s.%(name)s %(val)d %(time)d\n' %
                {
                    'prefix': self.config.get("Graphite", "YAStatsdPrefix"),
                    'name': 'gauges.count',
                    'val': len(stats.gauges_count),
                    'time': time
                })



from setuptools import setup

setup(name='yastatsd',
    version='0.1.5',
    description='Yet Another Statsd Server',
    author='Gregory Haynes',
    author_email='greg@greghaynes.net',
    url='http://github.com/greghaynes/YAStatsd',
    license='MIT',
    install_requires=['Twisted>=11.0.0'],
    packages=['yastatsd', 'yastatsd.backends'],
    scripts=['scripts/yastatsd']
    )
